import string

try:
    import secrets
except ImportError:
    import random as secrets


def gen_password(length=20, alphabet=None):
    if not alphabet:
        alphabet = string.ascii_letters + string.digits

    password = "".join([secrets.choice(alphabet) for _ in range(length)])
    print(password)


if __name__ == "__main__":
    gen_password(length=80)



import os

from .base import BaseFlowStep


class StepMakeDirectoryForBackup(BaseFlowStep):
    def run(self, stat_entry, dry_run=False):

        res = self._render_parameter("generation_mask")
        res = os.path.normpath(os.path.abspath(res))

        if not dry_run:
            os.makedirs(res)

        self.additional_output_context = {"result": res}

        return super().run(dry_run)

    @classmethod
    def step_name(cls):
        return "mkdir_for_backup"



import os
import re

from .base import DryRunExecutionError, time_interval
from .s3_steps_shared import StepS3FileBaseUploader, UploadTarget
from .s3boto_client import S3BasicBotoClient


class StepS3MultipartUploadWithRotation(StepS3FileBaseUploader):
    def __init__(self, **kwargs):
        self._first_uploads_key_name_per_files = {}
        super().__init__(**kwargs)

    def vote_for_flow_execution_skipping(self):
        self.logger.debug("checking if need skip flow execution for step '{}'".format(self.step_name))

        bucket_name = self.secret_context["bucket_name"]
        self.logger.debug("bucket_name: '{}'".format(bucket_name))

        raw_client = self._crete_s3_client()
        client = S3BasicBotoClient(self.logger, raw_client)

        prefix_in_bucket = self._render_parameter("prefix_in_bucket")
        self.logger.debug("prefix_in_bucket: '{}'".format(prefix_in_bucket))
        self.step_context["prefix_in_bucket"] = prefix_in_bucket

        target_prefix_in_bucket = self._render_parameter("target_prefix_in_bucket")
        self.logger.debug("target_prefix_in_bucket: '{}'".format(target_prefix_in_bucket))

        additional_context = {"target_prefix_in_bucket": target_prefix_in_bucket}
        upload_rules = self.mixed_context["upload_rules"]

        if not client.is_bucket_exists(bucket_name):
            self.logger.debug("Step can't be skipped, because bucket is not exists")
            return False

        for rule in upload_rules:
            self.logger.info("processing upload rule '{}'".format(rule["name"]))

            if not self._can_skip_execution_for_rule(
                client,
                bucket_name,
                rule,
                additional_context
            ):
                return False

        return True

    def run(self, stat_entry, dry_run=False):
        bucket_name = self.secret_context["bucket_name"]
        self.logger.debug("bucket_name: '{}'".format(bucket_name))

        if dry_run:
            pattern = re.compile(StepS3FileBaseUploader.S3_BUCKET_NAME_REGEX)
            if not pattern.match(bucket_name):
                raise DryRunExecutionError(
                    "base bucket name should match regex '{}'".format(StepS3FileBaseUploader.S3_BUCKET_NAME_REGEX)
                )

        region = self.secret_context["region"]

        raw_client = self._crete_s3_client()
        client = S3BasicBotoClient(self.logger, raw_client)

        prefix_in_bucket = self._render_parameter("prefix_in_bucket")
        self.logger.debug("prefix_in_bucket: '{}'".format(prefix_in_bucket))
        self.step_context["prefix_in_bucket"] = prefix_in_bucket

        target_prefix_in_bucket = self._render_parameter("target_prefix_in_bucket")
        self.logger.debug("target_prefix_in_bucket: '{}'".format(target_prefix_in_bucket))

        targets = self._get_upload_targets()

        additional_context = {"target_prefix_in_bucket": target_prefix_in_bucket}
        upload_rules = self.mixed_context["upload_rules"]

        if dry_run:
            self.logger.debug("checking that bucket exists to perform dry run")
            client.is_bucket_exists(bucket_name)
            return super().run(dry_run)

        if not client.is_bucket_exists(bucket_name):
            self.logger.info("creating bucker '{}'".format(bucket_name))
            client.create_bucket(bucket_name, region=region)

        targets = self._get_real_source_file_names_for_targets(targets)

        self.logger.info("going to upload these files:\n\t{}".format(targets))

        for rule in upload_rules:
            self.logger.info("processing upload rule '{}'".format(rule["name"]))
            self._upload_for_rule(
                stat_entry,
                client,
                bucket_name,
                rule,
                targets,
                additional_context
            )

        uploaded_size_metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_UPLOADED_SIZE
        )

        upload_time_metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_TRANSMISSION_TIME
        )

        if uploaded_size_metric.value and upload_time_metric.value:
            transmission_speed_metric = self._get_metric_by_name(
                stat_entry,
                StepS3FileBaseUploader.METRIC_TRANSMISSION_SPEED,
                units_name="MiB/s"
            )

            if upload_time_metric.value:
                transmission_speed_metric.value = round(uploaded_size_metric.value / upload_time_metric.value, 2)
            else:
                transmission_speed_metric.value = "N/A"

        if uploaded_size_metric.value:
            uploaded_size_metric.value = round(uploaded_size_metric.value, 2)

        if upload_time_metric.value:
            upload_time_metric.value = round(upload_time_metric.value, 3)

        return super().run(dry_run)

    def _get_upload_targets(self):
        raw_data = self.mixed_context["source_files"]

        res = []
        for raw_item in raw_data:
            res_item = UploadTarget()
            res_item.source_file = raw_item["source_file"]
            res_item.add_dedup_tag = raw_item.get("add_dedup_tag", False)

            res.append(res_item)

        return res

    def _can_skip_execution_for_rule(
        self,
        basic_client,
        bucket_name,
        rule,
        additional_context=None
    ):
        destination_prefix = self._render_result(rule["destination_prefix"], additional_context)
        self.logger.debug("destination_prefix: '{}'".format(destination_prefix))

        dedup_tag_name = self._render_result(rule["dedup_tag_name"], additional_context)
        dedup_tag_value = self._render_result(rule.get("dedup_tag_value"), additional_context)

        self.logger.debug(
            "looking for objects with tag '{}' and tag value '{}'".format(
                dedup_tag_name,
                dedup_tag_value
            )
        )
        tagged_key = self._get_tagged_object_key(
            basic_client,
            bucket_name,
            destination_prefix,
            dedup_tag_name,
            dedup_tag_value
        )
        self.logger.debug("tagged_key = '{}'".format(tagged_key))

        if tagged_key:
            self.logger.info("tag for iteration already exists: '{}'".format(tagged_key))
            return True

        return False

    def _upload_for_rule(
        self,
        stat_entry,
        basic_client,
        bucket_name,
        rule,
        upload_targets,
        additional_context=None
    ):
        destination_prefix = self._render_result(rule["destination_prefix"], additional_context)
        self.logger.debug("destination_prefix: '{}'".format(destination_prefix))

        # TODO: rename in configuration file
        dedup_tag_name = self._render_result(rule["dedup_tag_name"], additional_context)
        dedup_tag_value = self._render_result(rule.get("dedup_tag_value"), additional_context)
        marking_tags = {
            dedup_tag_name: dedup_tag_value
        }

        self.logger.debug(
            "looking for objects with tag '{}' and tag value '{}'".format(
                dedup_tag_name,
                dedup_tag_value
            )
        )
        tagged_key = self._get_tagged_object_key(
            basic_client,
            bucket_name,
            destination_prefix,
            dedup_tag_name,
            dedup_tag_value
        )
        self.logger.debug("tagged_key = '{}'".format(tagged_key))

        if tagged_key:
            self.logger.info("tag for iteration already exists: '{}'".format(tagged_key))
            return

        existing_files_for_rule = self._load_already_existing_files_for_rule(
            basic_client,
            bucket_name,
            destination_prefix
        )

        for upload_target in upload_targets:
            assert os.path.exists(upload_target.os_file_name)

            dest_key_name = os.path.join(destination_prefix, os.path.basename(upload_target.os_file_name))
            dest_key_name = str(dest_key_name).replace("\\", "/")

            self.logger.info("dest_key_name: '{}'".format(dest_key_name))

            first_upload_key_name = self._first_uploads_key_name_per_files.get(upload_target.os_file_name, None)
            self.logger.debug(
                "first_upload_key_name: '{}' for file '{}'".format(
                    first_upload_key_name,
                    upload_target.os_file_name
                )
            )

            if not first_upload_key_name:
                self.logger.info("no previous uploads available - FRESH UPLOAD")
                self.logger.info("bucket_name: '{}', dest_key_name: '{}'".format(bucket_name, dest_key_name))

                transmission_start_timestamp = self._get_current_timestamp()
                basic_client.upload_file(
                    bucket_name,
                    dest_key_name,
                    upload_target.os_file_name
                )
                transmission_end_timestamp = self._get_current_timestamp()

                metric = self._get_metric_by_name(
                    stat_entry,
                    StepS3FileBaseUploader.METRIC_UPLOADED_OBJECTS_COUNT,
                    initial_value=0,
                    units_name="items"
                )
                metric.increment(1)

                metric = self._get_metric_by_name(
                    stat_entry,
                    StepS3FileBaseUploader.METRIC_UPLOADED_SIZE,
                    initial_value=0.0,
                    units_name="MiB"
                )
                size_in_mibs = self._get_file_size_in_mibs(upload_target.os_file_name)
                metric.increment(size_in_mibs)

                metric = self._get_metric_by_name(
                    stat_entry,
                    StepS3FileBaseUploader.METRIC_TRANSMISSION_TIME,
                    initial_value=0.0,
                    units_name="seconds"
                )
                metric.increment(time_interval(transmission_start_timestamp, transmission_end_timestamp))

                self._first_uploads_key_name_per_files[upload_target.os_file_name] = dest_key_name
            else:
                self.logger.info(
                    "previous upload available in key '{}' will COPY to key '{}'".format(
                        first_upload_key_name,
                        dest_key_name
                    )
                )

                basic_client.copy_file_from_one_bucket_to_another(
                    bucket_name,
                    first_upload_key_name,
                    bucket_name,
                    dest_key_name
                )

                metric = self._get_metric_by_name(
                    stat_entry,
                    StepS3FileBaseUploader.METRIC_COPIED_OBJECTS_COUNT,
                    initial_value=0,
                    units_name="items"
                )
                metric.increment(1)

            if dest_key_name in existing_files_for_rule:
                existing_files_for_rule.remove(dest_key_name)

            if upload_target.add_dedup_tag:
                basic_client.set_object_tags(bucket_name, dest_key_name, marking_tags)

        self._remove_files_existing_for_rule(stat_entry, basic_client, bucket_name, existing_files_for_rule)

    def _load_already_existing_files_for_rule(self, basic_client, bucket_name, destination_prefix):
        self.logger.debug("checking for files that already exists in bucket")
        existing_files_for_rule = basic_client.list_files_in_folder(bucket_name, destination_prefix)
        self.logger.debug("existing_files_for_rule: {}".format(existing_files_for_rule))

        return existing_files_for_rule

    def _remove_files_existing_for_rule(self, stat_entry, basic_client, bucket_name, existing_files_for_rule):
        existing_files_base_names = [os.path.basename(item) for item in existing_files_for_rule]
        self.logger.info("some files already exists in folder for rule: {}".format(existing_files_base_names))

        for existing_file in existing_files_for_rule:
            self.logger.info("removing item '{}'".format(existing_file))
            basic_client.delete_object(bucket_name, existing_file)

            metric = self._get_metric_by_name(
                stat_entry,
                StepS3FileBaseUploader.METRIC_DELETED_OBJECTS_COUNT,
                initial_value=0,
                units_name="items"
            )
            metric.increment(1)

    @classmethod
    def step_name(cls):
        return "s3_multipart_upload_with_rotation"



import os

from jinja2 import BaseLoader, Environment, StrictUndefined


def jinja2_custom_filter_extract_year_four_digits(value):
    return value.strftime("%Y")


def jinja2_custom_filter_extract_month_two_digits(value):
    return value.strftime("%m")


def jinja2_custom_filter_extract_day_two_digits(value):
    return value.strftime("%d")


def create_rendering_environment():

    env = Environment(loader=BaseLoader, undefined=StrictUndefined)

    env.filters["extract_year_four_digits"] = jinja2_custom_filter_extract_year_four_digits
    env.filters["extract_month_two_digits"] = jinja2_custom_filter_extract_month_two_digits
    env.filters["extract_day_two_digits"] = jinja2_custom_filter_extract_day_two_digits
    env.filters["base_name"] = os.path.basename

    return env



import subprocess


class ThirdPartyCommandsExecutor(object):
    @staticmethod
    def execute(command):
        result = subprocess.run(command, stdout=subprocess.PIPE, stdin=subprocess.PIPE)

        result.stdout = result.stdout if result.stdout is not None else bytes()
        result.stderr = result.stderr if result.stderr is not None else bytes()

        return result



import codecs
import datetime
import hashlib
import os

from .base import BaseFlowStep, DryRunExecutionError


class StepCalculateFileHashAndSaveToFile(BaseFlowStep):
    def run(self, stat_entry, dry_run=False):
        input_file_name = self._render_parameter("input_file_name")
        self.step_context["input_file_name"] = input_file_name

        output_file_name = self._render_parameter("output_file_name")
        self.step_context["output_file_name"] = output_file_name

        hash_type = self.step_context["hash_type"]
        algorithms_available = [str(item).lower() for item in hashlib.algorithms_available]
        self.logger.debug(f"algorithms_available: {algorithms_available}")

        if hash_type not in algorithms_available:
            raise DryRunExecutionError(f"unsupported hash type '{hash_type}'")

        if not dry_run:
            self.logger.info(f"calculating hash ('{hash_type}') for '{input_file_name}'")

            hashing_begin_timestamp = datetime.datetime.utcnow()
            hash_value = self._hash_file(input_file_name, hash_type)
            hashing_end_timestamp = datetime.datetime.utcnow()

            metric = self._get_metric_by_name(stat_entry, "Hashed File")
            metric.value = os.path.basename(input_file_name)

            metric = self._get_metric_by_name(stat_entry, "Hash Type")
            metric.value = os.path.basename(hash_type)

            metric = self._get_metric_by_name(stat_entry, "File Size", units_name="MiB")
            size_in_mibs = self._get_file_size_in_mibs(input_file_name)
            metric.value = f"{size_in_mibs:.2f}"

            metric = self._get_metric_by_name(stat_entry, "Hash Speed", units_name="MiB/s")
            spent_time = (hashing_end_timestamp - hashing_begin_timestamp).total_seconds()

            megs_per_second = (size_in_mibs / spent_time) if spent_time else "N/A"

            if megs_per_second != "N/A":
                metric.value = f"{megs_per_second:.2f}"

            output_data = f"{hash_value} *{os.path.basename(input_file_name)}\n"
            self._save_data(output_file_name, output_data)

        return super().run(dry_run)

    @staticmethod
    def _save_data(file_name, data, codepage="utf-8"):
        with codecs.open(file_name, "w", codepage) as output_file:
            output_file.write(data)

    @staticmethod
    def _hash_file(file_name, hash_type):
        BLOCKSIZE = 65536

        hasher = hashlib.new(hash_type)
        with open(file_name, "rb") as afile:
            buf = afile.read(BLOCKSIZE)
            while len(buf) > 0:
                hasher.update(buf)
                buf = afile.read(BLOCKSIZE)

        return str(hasher.hexdigest())

    @classmethod
    def step_name(cls):
        return "calculate_file_hash_and_save_in_file"



import os

from .base import BaseFlowStep
from .shared import ThirdPartyCommandsExecutor


class StepMakePgDatabaseWinBackup(BaseFlowStep):
    def run(self, stat_entry, dry_run=False):
        os.environ["PGPASSWORD"] = self._render_parameter("db_password")

        backup_file_name = self._render_parameter("backup_file_name")
        self.step_context["backup_file_name"] = backup_file_name

        command = self._render_parameter("command_template")
        self.step_context["command"] = command

        dry_run_command = self._render_parameter("dry_run_command")
        self.step_context["dry_run_command"] = dry_run_command

        if not dry_run:
            self.logger.info("Making backup of Firebird database")
            self.logger.debug("going to execute: {}".format(command))
            result = ThirdPartyCommandsExecutor.execute(command)
        else:
            self.logger.debug("going to execute: {}".format(dry_run_command))
            result = ThirdPartyCommandsExecutor.execute(dry_run_command)

        del os.environ["PGPASSWORD"]

        if not dry_run:
            backup_log_name = self._render_parameter("backup_log_name")
            self.logger.debug(f"Saving log file from PG backup tool into {backup_log_name}")
            self._save_backup_log(backup_log_name, result.stderr)

            result.check_returncode()

        return super().run(dry_run)

    @staticmethod
    def _save_backup_log(backup_log_name: str, content: bytes):
        with open(backup_log_name, "wb") as output_file:
            output_file.write(content)

    @classmethod
    def step_name(cls):
        return "pg_win_backup"



import argparse
import os
import re

from .base import DryRunExecutionError, time_interval
from .s3_steps_shared import StepS3FileBaseUploader, UploadTarget
from .s3boto_client import S3BasicBotoClient


class StepS3StrictUploader(StepS3FileBaseUploader):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._execution_suffix = None

    def run(self, stat_entry, dry_run=False):
        self._execution_suffix = self._get_upload_suffix()
        self.logger.debug("execution_suffix: {}".format(self._execution_suffix))
        self.step_context["execution_suffix"] = self._execution_suffix

        bucket_name = self.secret_context["bucket_name"]
        self.logger.debug("bucket_name: '{}'".format(bucket_name))

        if dry_run:
            pattern = re.compile(StepS3FileBaseUploader.S3_BUCKET_NAME_REGEX)
            if not pattern.match(bucket_name):
                raise DryRunExecutionError(
                    "base bucket name should match regex '{}'".format(StepS3FileBaseUploader.S3_BUCKET_NAME_REGEX)
                )

        region = self.secret_context["region"]

        raw_client = self._crete_s3_client()
        client = S3BasicBotoClient(self.logger, raw_client)

        prefix_in_bucket = self._render_parameter("prefix_in_bucket")
        self.logger.debug("prefix_in_bucket: '{}'".format(prefix_in_bucket))
        self.step_context["prefix_in_bucket"] = prefix_in_bucket

        target_prefix_in_bucket = self._render_parameter("target_prefix_in_bucket")
        self.logger.debug("target_prefix_in_bucket: '{}'".format(target_prefix_in_bucket))

        uploads = self._get_uploads()
        self.logger.debug("uploads: {}".format(uploads))

        # additional_context = {"target_prefix_in_bucket": target_prefix_in_bucket}
        # uploads = self.mixed_context["uploads"]

        if dry_run:
            self.logger.debug("checking that bucket exists to perform dry run")
            client.is_bucket_exists(bucket_name)
            return super().run(dry_run)

        if not client.is_bucket_exists(bucket_name):
            self.logger.info("creating bucker '{}'".format(bucket_name))
            client.create_bucket(bucket_name, region=region)

        uploads = self._get_real_source_file_names_for_targets(uploads)
        self.logger.info("going to upload these files:\n\t{}".format(uploads))

        for upload_target in uploads:
            self.logger.info("processing upload rule '{}'".format(upload_target))
            self._upload_file(
                stat_entry,
                client,
                bucket_name,
                target_prefix_in_bucket,
                upload_target
            )

        uploaded_size_metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_UPLOADED_SIZE
        )

        upload_time_metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_TRANSMISSION_TIME
        )

        if uploaded_size_metric.value and upload_time_metric.value:
            transmission_speed_metric = self._get_metric_by_name(
                stat_entry,
                StepS3FileBaseUploader.METRIC_TRANSMISSION_SPEED,
                units_name="MiB/s"
            )

            if upload_time_metric.value:
                transmission_speed_metric.value = round(uploaded_size_metric.value / upload_time_metric.value, 2)
            else:
                transmission_speed_metric.value = "N/A"

        if uploaded_size_metric.value:
            uploaded_size_metric.value = round(uploaded_size_metric.value, 2)

        if upload_time_metric.value:
            upload_time_metric.value = round(upload_time_metric.value, 3)

        return super().run(dry_run)

    def _get_uploads(self):
        raw_data = self.mixed_context["uploads"]

        res = []
        for raw_item in raw_data:
            res_item = UploadTarget()
            res_item.source_file = raw_item["source_file"]
            res.append(res_item)

        return res

    def _upload_file(
        self,
        stat_entry,
        basic_client,
        bucket_name,
        target_prefix_in_bucket,
        upload_target
    ):
        assert os.path.exists(upload_target.os_file_name)

        dest_key_name = os.path.join(target_prefix_in_bucket, os.path.basename(upload_target.os_file_name))
        dest_key_name = str(dest_key_name).replace("\\", "/")

        self.logger.info("bucket_name: '{}', dest_key_name: '{}'".format(bucket_name, dest_key_name))

        transmission_start_timestamp = self._get_current_timestamp()
        basic_client.upload_file(
            bucket_name,
            dest_key_name,
            upload_target.os_file_name
        )
        transmission_end_timestamp = self._get_current_timestamp()

        metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_UPLOADED_OBJECTS_COUNT,
            initial_value=0,
            units_name="items"
        )
        metric.increment(1)

        metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_UPLOADED_SIZE,
            initial_value=0.0,
            units_name="MiB"
        )
        size_in_mibs = self._get_file_size_in_mibs(upload_target.os_file_name)
        metric.increment(size_in_mibs)

        metric = self._get_metric_by_name(
            stat_entry,
            StepS3FileBaseUploader.METRIC_TRANSMISSION_TIME,
            initial_value=0.0,
            units_name="seconds"
        )
        metric.increment(time_interval(transmission_start_timestamp, transmission_end_timestamp))

    def _get_upload_suffix(self):
        unknown_args = self.rendering_context.unknown_args
        self.logger.debug("unknown_args: {}".format(unknown_args))

        parser = self._init_arg_parser()
        args, _ = parser.parse_known_args(unknown_args)
        upload_suffix = None
        if args.upload_suffix:
            upload_suffix = str(args.upload_suffix).strip()
            upload_suffix = None if not upload_suffix else upload_suffix

        if upload_suffix and (upload_suffix[0] != "-"):
            upload_suffix = "-" + upload_suffix

        return upload_suffix

    @staticmethod
    def _init_arg_parser():
        parser = argparse.ArgumentParser()

        parser.add_argument(
            "--upload-suffix",
            action="store",
            help="Upload suffix for strict upload to S3"
        )

        return parser

    @classmethod
    def step_name(cls):
        return "step_s3_strict_upload"



class StatMetricEntry(object):
    def __init__(self, metric_name, initial_value=None, units_name=None):
        assert str(metric_name).strip()
        self._metric_name = metric_name
        self._value = initial_value
        self.units_name = units_name

    @property
    def metric_name(self):
        return self._metric_name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    def increment(self, delta):
        self._value += delta


class MetricsHolder(object):
    def __init__(self):
        self._metrics = {}

    def get_metric(self, metric_name, initial_value=None, units_name=None):
        res = self._metrics.get(metric_name)

        if res is None:
            res = StatMetricEntry(metric_name=metric_name, initial_value=initial_value, units_name=units_name)
            self._metrics[metric_name] = res

        return res

    def get_all_metrics(self):
        return list(self._metrics.keys())

    def is_empty(self):
        return True if not self._metrics else False


class StepExecutionStatisticEntry(object):
    def __init__(
        self,
        step_name,
        step_human_readable_name=None,
        execution_start_timestamp=None,
        execution_end_timestamp=None,
        metrics=None
    ):
        self.step_name = step_name
        self.step_human_readable_name = step_human_readable_name
        self.execution_start_timestamp = execution_start_timestamp
        self.execution_end_timestamp = execution_end_timestamp
        self.metrics = MetricsHolder() if metrics is None else metrics



import datetime
import os


class DryRunExecutionError(Exception):
    pass


class TransmissionError(Exception):
    pass


class WrongParameterTypeError(Exception):
    pass


class StepContextData(object):
    def __init__(self):
        self.name = None
        self.description = None


def pretty_time_delta(seconds):
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(seconds)

    days, seconds = divmod(seconds, 86400)
    days = int(days)

    hours, seconds = divmod(seconds, 3600)
    hours = int(hours)

    minutes, seconds = divmod(seconds, 60)
    minutes = int(minutes)

    seconds = round(seconds, 3)

    if days > 0:
        return "{}{}d{}h{}m{}s".format(sign_string, days, hours, minutes, seconds)
    elif hours > 0:
        return "{}{}h{}m{}s".format(sign_string, hours, minutes, seconds)
    elif minutes > 0:
        return "{}{}m{}s".format(sign_string, minutes, seconds)
    else:
        return "{}{:.3f}s".format(sign_string, seconds)


def time_interval(timestamp_start, timestamp_end):
    time_elapsed = (timestamp_end - timestamp_start)
    res = float(time_elapsed.total_seconds())

    res += round(time_elapsed.microseconds / 1000000, 3)
    return res


class BaseFlowStep(object):
    BYTES_IN_MEGABYTE = 1024 * 1024

    def __init__(
        self,
        logger,
        rendering_context,
        step_context,
        secret_context,
        rendering_environment,
    ):
        self.logger = logger
        self.rendering_context = rendering_context
        self.step_context = step_context
        self.rendering_environment = rendering_environment
        self.secret_context = secret_context
        self.additional_output_context = None

    @classmethod
    def step_name(cls):
        pass

    @property
    def mixed_context(self):
        return self._get_mixed_context()

    def run(self, stat_entry, dry_run=False):
        output_variables = self._generate_output_variables()
        self.logger.debug("output_variables: {}".format(output_variables))

        return output_variables

    def vote_for_flow_execution_skipping(self):
        return None

    def _render_parameter(self, parameter_name, context=None):
        if not context:
            context = self.mixed_context

        template = context[parameter_name]
        self.logger.debug("'{}'@template: '{}'".format(parameter_name, template))

        res = self._render_result(template)
        self.logger.debug("'{}'@value: '{}'".format(parameter_name, res))

        return res

    def _get_mixed_context(self):
        mixed_context = self.rendering_context.to_context()
        mixed_context = {**mixed_context, **self._get_step_context()}

        return mixed_context

    def _get_step_context(self):
        return {**self.secret_context, **self.step_context}

    def _render_result(self, template, additional_context=None):
        if not template:
            return ""

        mixed_context = self.mixed_context

        if additional_context:
            mixed_context = {**mixed_context, **additional_context}

        return self._render_from_template_and_context(template, mixed_context)

    def _render_from_template_and_context(self, template, context):
        jinja2_template = self.rendering_environment.from_string(template)
        return jinja2_template.render(**context)

    def _generate_output_variables(self):
        res = dict()

        if "generates" not in self.step_context:
            return res

        for requested_value_name, requested_value_template in self.step_context["generates"].items():
            res[requested_value_name] = self._render_result(requested_value_template, self.additional_output_context)

        return res

    def _get_metric_by_name(self, stat_entry, metric_name, initial_value=None, units_name=None):
        return stat_entry.metrics.get_metric(metric_name, initial_value=initial_value, units_name=units_name)

    def _get_file_size_in_mibs(self, file_name):
        file_size = os.path.getsize(file_name)

        size_in_megs = (file_size / BaseFlowStep.BYTES_IN_MEGABYTE)
        return size_in_megs

    def _get_current_timestamp(self):
        return datetime.datetime.utcnow()



from .base import BaseFlowStep, time_interval
from .shared import ThirdPartyCommandsExecutor


class StepCompressFileWith7Z(BaseFlowStep):

    def run(self, stat_entry, dry_run=False):
        output_archive_name = self._render_parameter("output_archive_name")
        self.step_context["output_archive_name"] = output_archive_name

        command = self._render_parameter("command_template")
        self.step_context["command"] = command

        dry_run_command = self._render_parameter("dry_run_command")
        self.step_context["dry_run_command"] = dry_run_command

        timestamp_execution_start = self._get_current_timestamp()
        if not dry_run:
            self.logger.info("Compressing file with 7Z archive")
            self.logger.debug("going to execute: {}".format(command))
            result = ThirdPartyCommandsExecutor.execute(command)
        else:
            self.logger.debug("going to execute: {}".format(dry_run_command))
            result = ThirdPartyCommandsExecutor.execute(dry_run_command)

        self.logger.info("return code: {}".format(result.returncode))

        timestamp_execution_end = self._get_current_timestamp()

        if not dry_run:
            self.logger.info("stderr:\n{}".format(result.stderr.decode("utf-8")))
            self.logger.info("stdout:\n{}".format(result.stdout.decode("utf-8")))

        if not dry_run:
            result.check_returncode()

            size_in_mibs = self._get_file_size_in_mibs(output_archive_name)
            spen_time = time_interval(timestamp_execution_start, timestamp_execution_end)
            speed_in_mibs = (size_in_mibs / spen_time) if spen_time else "N/A"

            self._get_metric_by_name(
                stat_entry,
                "Compressed Size",
                initial_value=round(size_in_mibs, 2),
                units_name="MiB"
            )

            self._get_metric_by_name(
                stat_entry,
                "Compression Speed",
                initial_value=round(speed_in_mibs, 2),
                units_name="MiB/s"
            )

        return super().run(dry_run)

    @classmethod
    def step_name(cls):
        return "7z_compress"



from .factory import create_steps_factory

from .step_calculate_file_hash_and_save_to_file import StepCalculateFileHashAndSaveToFile
from .step_compress_file_with_7z import StepCompressFileWith7Z
from .step_make_directory_for_backup import StepMakeDirectoryForBackup
from .step_make_firebird_database_backup import StepMakeFirebirdDatabaseBackup
from .step_s3_multipart_upload_with_rotation import StepS3MultipartUploadWithRotation
from .step_validate_7z_archive import StepValidate7ZArchive
from .step_s3_strict_uploader import StepS3StrictUploader



import datetime
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import mimetypes
import os
import socket

import boto3
from botocore.exceptions import ClientError
from yabtool.shared.base import AttrsToStringMixin
from yabtool.shared.jinja2_helpers import create_rendering_environment
from yabtool.supported_steps.base import pretty_time_delta, time_interval
from yabtool.version import __version__


class DataForEmailSending(AttrsToStringMixin):
    def __init__(self):
        self.subject = None
        self.sender = None
        self.to_recipients = None
        self.cc_recipients = None
        self.bcc_recipients = None
        self.plain_text_body = None
        self.html_body = None
        self.charset = None
        self.attachments = []


class EmailSender(object):

    def __init__(self, logger, notification_data):
        self.logger = logger
        self.notification_data = notification_data

    def send(self, data_for_sending: DataForEmailSending):
        connection = self.notification_data.get("connection")
        assert connection

        region = connection.get("region")
        assert region

        connection_data = self.notification_data.get("connection")
        assert connection_data

        aws_access_key_id = connection_data.get("aws_access_key_id")
        assert aws_access_key_id

        aws_secret_access_key = connection_data.get("aws_secret_access_key")
        assert aws_secret_access_key

        client = boto3.client(
            "ses",
            region_name=region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key
        )

        msg = MIMEMultipart()
        msg["Subject"] = data_for_sending.subject
        msg["From"] = data_for_sending.sender
        msg["To"] = "; ".join(data_for_sending.to_recipients)

        for attachment_file_path in data_for_sending.attachments:
            part = MIMEApplication(self._get_file_content(attachment_file_path))
            part.add_header("Content-Disposition", "attachment", filename=os.path.basename(attachment_file_path))
            mimetype = mimetypes.guess_type(attachment_file_path)
            part.add_header("Content-Type", "{}; charset=UTF-8".format(mimetype))

            msg.attach(part)

        if data_for_sending.html_body:
            part_html = MIMEText(data_for_sending.html_body, "html")
            msg.attach(part_html)
        else:
            part_plain = MIMEText(data_for_sending.plain_text_body, "plain")
            msg.attach(part_plain)

        try:
            self.logger.info("sending email to {}".format(data_for_sending.to_recipients))

            response = client.send_raw_email(
                RawMessage={"Data": msg.as_string()},
                Source=data_for_sending.sender,
                Destinations=data_for_sending.to_recipients
            )

        # Display an error if something goes wrong.
        except ClientError as e:
            self.logger.error("Error sending email: {}".format(e.response["Error"]["Message"]))
        except Exception as e:
            self.logger.exception("Error sending email: {}".format(e))
        else:
            self.logger.info("Email sent! Message ID: {}".format(response["MessageId"]))

    def _get_file_content(self, file_name):
        with open(file_name, "rb") as input_file:
            return input_file.read()


class EmailRenderer(object):
    DEFAULT_CHARSET = "utf-8"

    def __init__(self, logger):
        self.logger = logger
        self.flow_orchestrator = None
        self.notification_data = None
        self.succeeded = None
        self.exception = None
        # self.seconds_spent = None
        self.only_dry_run = None

        self.additional_variables = {}

        self.charset = EmailRenderer.DEFAULT_CHARSET

    def render(self):
        res = DataForEmailSending()
        res.charset = self.charset

        timestamp_end = datetime.datetime.utcnow()
        seconds_spent = time_interval(self.flow_orchestrator.backup_start_timestamp, timestamp_end)

        flow_context = self.flow_orchestrator.rendering_context

        dry_run_stat = self.flow_orchestrator.produce_exeuction_stat(self.flow_orchestrator.dry_run_statistics)
        active_run_stat = self.flow_orchestrator.produce_exeuction_stat(self.flow_orchestrator.active_run_statistics)

        dry_run_metrics_data_list = self.flow_orchestrator.produce_execution_metrics(
            self.flow_orchestrator.dry_run_statistics
        )
        dry_run_metrics = ""
        for step_name, metrics_data_item in dry_run_metrics_data_list:
            dry_run_metrics += "\nMetrics for '{}':\n{}".format(step_name, metrics_data_item)
        dry_run_metrics = str(dry_run_metrics).strip()

        active_run_metrics_data_list = self.flow_orchestrator.produce_execution_metrics(
            self.flow_orchestrator.active_run_statistics
        )
        active_run_metrics = ""
        for step_name, metrics_data_item in active_run_metrics_data_list:
            active_run_metrics += "\n\nMetrics for '{}':\n{}".format(step_name, metrics_data_item)
        active_run_metrics = str(active_run_metrics).strip()

        rendering_context = flow_context.to_context()

        new_values = {
            "backup_start_timestamp": self.flow_orchestrator.backup_start_timestamp,
            "flow_name": flow_context.flow_name,
            "host_name": socket.gethostname(),
            "flow_execution_succeeded": self.succeeded,
            "str_flow_execution_status": ("succeeded" if self.succeeded else "failed"),
            "time_spent": pretty_time_delta(seconds_spent),
            "flow_exception": str(self.exception),

            "dry_run_stat": dry_run_stat,
            "active_run_stat": active_run_stat,
            "only_dry_run": self.only_dry_run,

            "dry_run_metrics": dry_run_metrics,
            "active_run_metrics": active_run_metrics,
            "yabtool_version": __version__
        }

        rendering_context = {**rendering_context, **new_values}
        rendering_context = {**rendering_context, **self.additional_variables}

        res.sender = self.notification_data.get("sender")
        assert res.sender

        res.to_recipients = self.notification_data.get("to")
        assert res.to_recipients

        subject_template = self.notification_data.get("subject")
        assert subject_template

        plain_body_template = self.notification_data.get("body")
        assert plain_body_template

        res.subject, res.plain_text_body = self._render_plain_text_body(
            rendering_context,
            plain_body_template,
            subject_template
        )
        self.logger.info("plain_body:\n{}".format(res.plain_text_body))

        res.html_body = self._render_html_body(res.plain_text_body)

        return res

    def _render_plain_text_body(self, context, plain_body_template, subject_template):
        rendering_environment = create_rendering_environment()

        body_jinja2_template = rendering_environment.from_string(plain_body_template)
        subject_jinja2_template = rendering_environment.from_string(subject_template)

        plain_body = body_jinja2_template.render(**context)
        plain_subject = subject_jinja2_template.render(**context)

        return plain_subject, plain_body

    def _render_html_body(self, plain_text):
        return """<html>
            <head></head>
            <body>
              <pre>{}</pre>
            </body>
            </html>
        """.format(plain_text)



import copy
import os

import boto3
from yabtool.shared.base import AttrsToStringMixin

from .base import BaseFlowStep, TransmissionError


class UploadTarget(AttrsToStringMixin):
    def __init__(self):
        self.source_file = None
        self.add_dedup_tag = False
        self.os_file_name = None


class StepS3FileBaseUploader(BaseFlowStep):
    S3_BUCKET_NAME_REGEX = r"^[a-zA-Z0-9.\-_]{1,255}$"

    METRIC_UPLOADED_OBJECTS_COUNT = "Uploaded Objects"
    METRIC_UPLOADED_SIZE = "Uploaded Size"
    METRIC_TRANSMISSION_TIME = "Transmission Time"
    METRIC_TRANSMISSION_SPEED = "Transmission Speed"
    METRIC_COPIED_OBJECTS_COUNT = "Copied Objects Count"
    METRIC_DELETED_OBJECTS_COUNT = "Deleted Objects Count"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._client = None

    def _crete_s3_client(self):
        region = self.secret_context.get("region")
        self.logger.debug("S3 region: '{}'".format(region))

        return boto3.client(
            "s3",
            region_name=region,
            aws_access_key_id=self.secret_context["aws_access_key_id"],
            aws_secret_access_key=self.secret_context["aws_secret_access_key"]
        )

    def _get_tagged_object_key(
        self,
        basic_client,
        bucket_name,
        destination_prefix,
        validation_tag_name,
        validation_tag_value
    ):
        objects_for_prefix = basic_client.list_files_in_folder(bucket_name, destination_prefix)

        for object_key in objects_for_prefix:
            object_tags = basic_client.get_object_tags(bucket_name, object_key)

            self.logger.debug("tags for key '{}': {}".format(object_key, object_tags))

            if validation_tag_name not in object_tags:
                continue

            if object_tags[validation_tag_name] == validation_tag_value:
                return object_key

        return None

    def _get_real_source_file_names_for_targets(self, targets):
        res = []

        for target in targets:
            new_target = copy.deepcopy(target)

            source_file_name = new_target.source_file
            source_file_name = self._render_result(source_file_name)
            self.logger.debug("source_file: {}".format(source_file_name))
            if not os.path.exists(source_file_name):
                msg = "Can't find source file '{}'".format(source_file_name)
                self.logger.error(msg)
                raise TransmissionError(msg)

            new_target.os_file_name = source_file_name
            res.append(new_target)

        return res



class AttrsToStringMixin(object):
    def __repr__(self):
        data = ["{}={}".format(k, v) for k, v in self.__dict__.items()]
        return "{}({})".format(self.__class__.__name__, ", ".join(data))

    def __str__(self):
        return self.__repr__()



__version__ = "0.0.9.4"
__app_name__ = "yabtool"
__human_readable_app_name__ = "Yet Another Backup Tool"
__package_description__ = (
    "Yet Another Backup Tool with controlled flow and archives rolling"
)



import datetime

from .supported_steps.base import pretty_time_delta, time_interval
from .yabtool_application import YabtoolApplication


if __name__ == "__main__":
    timestamp_start = datetime.datetime.utcnow()

    app = YabtoolApplication()
    app.run()

    timestamp_end = datetime.datetime.utcnow()
    seconds_spent = time_interval(timestamp_start, timestamp_end)

    app.logger.info("app finished @ {}".format(pretty_time_delta(seconds_spent)))



import codecs
import copy
import datetime
import os
import uuid

import terminaltables
from yabtool.shared.jinja2_helpers import create_rendering_environment
from yaml import safe_load

from .supported_steps import create_steps_factory
from .supported_steps.base import pretty_time_delta, time_interval
from .yabtool_stat import StepExecutionStatisticEntry

DEFAULT_CONFIG_RELATIVE_NAME = "./config/config.yaml"


class ConfigurationValidationException(BaseException):
    pass


class RenderingContext(object):
    def __init__(self):
        self.config_file_name = None
        self.secrets_file_name = None

        self.config_context = dict()
        self.secrets_context = dict()
        self.target_name = None
        self.flow_name = None

        self.basic_values = dict()
        self.previous_steps_values = list()
        self.temporary_folder = None
        self.root_temporary_folder = None

        self.remove_temporary_folder = None
        self.perform_dry_run = None
        self.unknown_args = None

    def to_context(self):
        res = self.basic_values
        for item in self.previous_steps_values:
            res = {**res, **item}

        return res


class YabtoolFlowOrchestrator(object):
    def __init__(self, logger):
        self.rendering_context = RenderingContext()
        self.logger = logger
        self._steps_factory = None
        self._backup_start_timestamp = datetime.datetime.utcnow()
        self._skip_flow_execution_voting_result = None
        self.skip_voting_enabled = True

        self.dry_run_statistics = []
        self.active_run_statistics = []

    def initialize(self, args, unknown_args):
        self.rendering_context.config_file_name = self._get_config_file_name(args)
        self.logger.debug(
            "config_file_name: '{}'".format(self.rendering_context.config_file_name)
        )

        if not self.rendering_context.config_file_name:
            raise ConfigurationValidationException("No configuration file specified")

        if not os.path.exists(self.rendering_context.config_file_name):
            raise ConfigurationValidationException(
                "Configuration file does not exists. Path: '{}'".format(
                    self.rendering_context.config_file_name
                )
            )

        self.rendering_context.config_context = self._load_yaml_file(
            self.rendering_context.config_file_name
        )

        self.rendering_context.unknown_args = unknown_args

        self.rendering_context.secrets_file_name = self._get_secrets_file_name(args)
        if not self.rendering_context.secrets_file_name:
            raise ConfigurationValidationException(
                "Secrets file is not specified"
            )

        if not os.path.exists(self.rendering_context.secrets_file_name):
            raise ConfigurationValidationException(
                "Secrets file does not exists. Path: '{}'".format(
                    self.rendering_context.secrets_file_name
                )
            )

        self.logger.debug("loading secrets from: '{}'".format(self.rendering_context.secrets_file_name))
        self.rendering_context.secrets_context = self._load_yaml_file(
            self.rendering_context.secrets_file_name
        )

        self.rendering_context.target_name = self._get_target_name(args)
        self.logger.debug("target_name: '{}'".format(self.target_name))

        self.rendering_context.flow_name = self._get_flow_name(args)
        self.logger.debug("flow_name: '{}'".format(self.rendering_context.flow_name))

        self._override_config_parameters_with_secrets()

        self.rendering_context.remove_temporary_folder = self.config_context["parameters"]["remove_temporary_folder"]
        self.rendering_context.perform_dry_run = self.config_context["parameters"]["perform_dry_run"] or args.dry_run

        self.rendering_context.root_temporary_folder = self._get_temporary_folder(args)

        self.rendering_context.temporary_folder = os.path.join(
            self.rendering_context.root_temporary_folder,
            self._create_folder_name_for_execution(),
        )

        os.makedirs(self.rendering_context.temporary_folder)

        self.logger.debug(
            "temporary_folder: '{}'".format(self.rendering_context.temporary_folder)
        )

        self._steps_factory = create_steps_factory()

        self.rendering_context.basic_values = self._init_basic_values()
        self.logger.debug(
            "basic_values: {}".format(self.rendering_context.basic_values)
        )

        return True

    def dry_run(self):
        self.logger.warning("performing dry run")
        if self.rendering_context.perform_dry_run:
            self._run(dry_run=True)

    def run(self):
        self.logger.warning("performing active run")
        self._run(dry_run=False)

    def print_stat(self):
        if self.dry_run_statistics:
            stat_data = self.produce_exeuction_stat(self.dry_run_statistics)
            self.logger.info("{}:\n{}".format("Dry run statistics:", stat_data))
            metrics_data_list = self.produce_execution_metrics(self.dry_run_statistics)
            for step_name, metrics_data_item in metrics_data_list:
                self.logger.info("Metrics for '{}':\n{}".format(step_name, metrics_data_item))

        if self.active_run_statistics:
            stat_data = self.produce_exeuction_stat(self.active_run_statistics)
            self.logger.info("{}:\n{}".format("Execution statistics:", stat_data))
            metrics_data_list = self.produce_execution_metrics(self.active_run_statistics)
            for step_name, metrics_data_item in metrics_data_list:
                self.logger.info("Metrics for '{}':\n{}".format(step_name, metrics_data_item))

        if (not self.dry_run_statistics) and (not self.active_run_statistics):
            self.logger.info("No execution statistics")

    def produce_exeuction_stat(self, stat_source):
        header = ["Step Name", "Exexcution start timestamp", "Execution end timestamp", "Time elapsed "]
        data = [header]

        total_time_elapsed_seconds = 0
        max_length = len(data[0]) if data else 0
        print_format = "%Y-%m-%d %H:%M:%S"
        for statistics_item in stat_source:
            # step_name = "{} ({})".format(statistics_item.step_human_readable_name, statistics_item.step_name)
            step_name = statistics_item.step_name
            time_elapsed_in_seconds = time_interval(
                statistics_item.execution_start_timestamp,
                statistics_item.execution_end_timestamp,
            )

            data_row = [
                step_name,
                statistics_item.execution_start_timestamp.strftime(print_format),
                statistics_item.execution_end_timestamp.strftime(print_format),
                pretty_time_delta(time_elapsed_in_seconds)
            ]

            data.append(data_row)
            max_length = max_length if max_length <= len(data_row) else len(data_row)

            total_time_elapsed_seconds += time_elapsed_in_seconds

        data_row = [""] * (max_length - 2)
        data_row.append("Total")
        data_row.append(pretty_time_delta(total_time_elapsed_seconds))
        data.append(data_row)

        table = terminaltables.AsciiTable(data)
        return table.table

    def produce_execution_metrics(self, stat_source):
        res = []
        for statistics_item in stat_source:
            step_name = "{} ({})".format(statistics_item.step_human_readable_name, statistics_item.step_name)

            metrics = statistics_item.metrics
            if metrics.is_empty():
                continue

            metrics_data = [["Name", "Value"]]

            for metric_name in metrics.get_all_metrics():
                metric = metrics.get_metric(metric_name)
                metrics_data.append(
                    [
                        metric.metric_name,
                        "{} {}".format(
                            metric.value,
                            metric.units_name
                        )
                    ]
                )

            table = terminaltables.AsciiTable(metrics_data)
            res_item = (step_name, table.table)
            res.append(res_item)

        return res

    def _get_config_file_name(self, args):
        config_file_name = args.config

        if not config_file_name:
            config_file_name = DEFAULT_CONFIG_RELATIVE_NAME
            src_path = os.path.abspath(os.path.dirname(__file__))
            assert os.path.exists(src_path)
            self.logger.debug("src_path: '{}'".format(src_path))

            config_file_name = os.path.join(src_path, config_file_name)
            config_file_name = os.path.abspath(config_file_name)
            config_file_name = os.path.normpath(config_file_name)

            if not os.path.exists(config_file_name):
                self.logger.error("can't find configuration file in '{}'".format(config_file_name))

            assert os.path.exists(config_file_name), config_file_name
        else:
            config_file_name = os.path.abspath(config_file_name)
            config_file_name = os.path.normpath(config_file_name)

        return config_file_name

    def _run(self, dry_run):
        assert self.rendering_context.flow_name

        if dry_run:
            self._skip_flow_execution_voting_result = None

        flow_data = self.rendering_context.config_context["flows"][self.flow_name]
        flow_description = flow_data["description"]

        self.logger.info("flow_description: '{}'".format(flow_description))

        self.rendering_context.previous_steps_values = []

        rendering_environment = create_rendering_environment()
        secret_targets_context = self.rendering_context.secrets_context["targets"][self.target_name]

        self._execute_steps(dry_run, flow_data, rendering_environment, secret_targets_context)

    def _execute_steps(self, dry_run, flow_data, rendering_environment, secret_targets_context):
        assert self._steps_factory

        statistics_list = self.dry_run_statistics if dry_run else self.active_run_statistics
        statistics_list.clear()

        if self._need_skip_voting():
            self.logger.warning("Want skip flow execution")
            return

        positive_votes_for_flow_execution_skipping = []
        for step_context in flow_data["steps"]:
            step_name = step_context["name"]
            step_human_readable_name = step_context.get("human_readable_name", step_name)

            step_description = step_context.get("description", "<no description>")

            self.logger.debug(
                "validating step '{}': {}".format(step_name, step_description)
            )

            if not self._steps_factory.is_step_known(step_name):
                raise ConfigurationValidationException(
                    "Unknown step '{}'".format(step_name)
                )

            if dry_run:
                self.logger.debug("performing dry run for step '{}'".format(step_name))
            else:
                self.logger.debug("performing active run for step '{}'".format(step_name))

            secret_context = dict()
            relative_secrets = step_context.get("relative_secrets", [])
            required_secrets = [step_name]
            required_secrets.extend(relative_secrets)

            for required_secret in required_secrets:
                if (
                    ("steps_configuration" in secret_targets_context) and  # noqa
                    (required_secret in secret_targets_context["steps_configuration"])
                ):
                    secret_context = {
                        **secret_context,
                        **secret_targets_context["steps_configuration"][required_secret]
                    }

            step_object = self._steps_factory.create_object(
                step_name,
                logger=self.logger,
                rendering_context=self.rendering_context,
                step_context=step_context,
                secret_context=secret_context,
                rendering_environment=rendering_environment,
            )

            if dry_run:
                self.logger.info("initializing dry run for step: '{}'".format(step_name))
                self.logger.debug("checking for decision for flow skipping")
                self._check_for_flow_execution_skipping(step_object, positive_votes_for_flow_execution_skipping)
            else:
                self.logger.info("initializing active run for step: '{}'".format(step_name))

            stat_entry = StepExecutionStatisticEntry(
                step_name=step_name,
                step_human_readable_name=step_human_readable_name,
                execution_start_timestamp=datetime.datetime.utcnow()
            )
            additional_variables = step_object.run(stat_entry, dry_run=dry_run)
            stat_entry.execution_end_timestamp = datetime.datetime.utcnow()

            statistics_list.append(stat_entry)

            self.logger.debug("additional_variables: {}".format(additional_variables))

            self.rendering_context.previous_steps_values.append(additional_variables)

        if dry_run and positive_votes_for_flow_execution_skipping:
            self.logger.info(
                "Flow execution can be SKIPPED.\n\tThese steps voted to skip flow execution: {}".format(
                    positive_votes_for_flow_execution_skipping
                )
            )
            self._skip_flow_execution_voting_result = True

    def _check_for_flow_execution_skipping(self, step_object, positive_votes_for_flow_execution_skipping):
        step_name = step_object.step_name()

        if self._skip_flow_execution_voting_result is not None:
            self.logger.debug(
                "decision for flow execution skipping already made. Can skip flow execution: {}".format(
                    self._skip_flow_execution_voting_result
                )
            )
            return

        vote = step_object.vote_for_flow_execution_skipping()
        if vote is None:
            self.logger.debug("step '{}' do not want to vote for flow execution skipping".format(step_name))
            return

        self.logger.debug("step '{}' voted for flow execution skipping".format(step_name))

        if not vote:
            self._skip_flow_execution_voting_result = False
            self.logger.debug("step '{}' voted against flow execution skipping".format(step_name))
        else:
            positive_votes_for_flow_execution_skipping.append(step_name)
            self.logger.debug("step '{}' voted for flow execution skipping".format(step_name))

    def _get_target_name(self, args):
        if args.target:
            return args.target

        assert self.rendering_context.secrets_context

        return self.rendering_context.secrets_context["defaults"]["target"]

    def _get_flow_name(self, args):
        if args.flow:
            return args.flow

        assert self.rendering_context.secrets_context
        assert self.rendering_context.target_name

        targets_context = self.rendering_context.secrets_context["targets"][
            self.rendering_context.target_name
        ]
        return targets_context["flow_type"]

    def _override_config_parameters_with_secrets(self):
        if "parameters" in self.secrets_context:
            override_params = self.secrets_context["parameters"]
            for key, value in override_params.items():
                self.config_context["parameters"][key] = value

        secret_targets_context = self.secrets_context["targets"][self.target_name]
        if "config_patch" in secret_targets_context:
            config_patch = secret_targets_context["config_patch"]

            patched_steps_count = 0
            for step_patch in config_patch["steps"]:
                step_patch_data = copy.deepcopy(step_patch)
                self.logger.debug("step_patch_data: {}".format(step_patch_data))
                del step_patch_data["name"]

                patched_steps_count += self._patch_step_in_flow(step_patch["name"], step_patch_data)

            if patched_steps_count:
                flow_data = self.rendering_context.config_context["flows"][self.flow_name]
                self.logger.debug("flow_data:\n{}".format(flow_data))

    def _patch_step_in_flow(self, step_name, step_patch_data):
        flow_data = self.rendering_context.config_context["flows"][self.flow_name]
        flow_steps = flow_data["steps"]

        patched_steps_count = 0
        flow_steps_count = len(flow_steps)
        for index in range(flow_steps_count):
            flow_step = flow_steps[index]
            if flow_step["name"] != step_name:
                continue

            self.logger.warning(
                "patching step '{}' for flow '{}' with:\n{}".format(
                    step_name,
                    self.flow_name,
                    step_patch_data
                )
            )
            flow_steps[index] = {**flow_step, **step_patch_data}
            patched_steps_count += 1

        return patched_steps_count

    def _get_temporary_folder(self, args):
        if args.temporary_folder:
            return args.temporary_folder

        temporary_folder = self.rendering_context.secrets_context["defaults"]["temporary_folder"]
        if temporary_folder:
            return temporary_folder

        return self.rendering_context.config_context["defaults"]["temporary_folder"]

    def _need_skip_voting(self):
        if not self.skip_voting_enabled:
            return False

        if (self._skip_flow_execution_voting_result is not None) and self._skip_flow_execution_voting_result:
            return True

        return False

    def _create_folder_name_for_execution(self):
        res = "{}_{}".format(self._backup_start_timestamp.isoformat(), str(uuid.uuid4()))
        res = res.replace(":", "")
        res = res.replace("-", "")
        return res

    @staticmethod
    def _load_yaml_file(file_name, codec="utf-8"):
        with codecs.open(file_name, "r", codec) as input_file:
            return safe_load(input_file.read())

    @staticmethod
    def _get_secrets_file_name(args):
        secrets_file_name = args.secrets
        if not secrets_file_name:
            return None

        secrets_file_name = os.path.abspath(secrets_file_name)
        secrets_file_name = os.path.normpath(secrets_file_name)

        return secrets_file_name

    def _init_basic_values(self):
        res = self._get_additional_rendering_variables()

        res["main_target_name"] = self.rendering_context.target_name
        res["week_day_short_name"] = self._backup_start_timestamp.strftime("%a")
        res["week_number"] = self._backup_start_timestamp.strftime("%U")
        res["month_short_name"] = self._backup_start_timestamp.strftime("%b")
        res["month_two_digit_number"] = self._backup_start_timestamp.strftime("%m")
        res["backup_start_timestamp"] = self._backup_start_timestamp
        res["flow_name"] = self.flow_name
        res["yabtool_exec_folder"] = self.rendering_context.temporary_folder

        res["current_year"] = self._backup_start_timestamp.strftime("%Y")
        res["current_month"] = res["month_two_digit_number"]
        res["current_day_of_month"] = self._backup_start_timestamp.strftime("%d")

        res["current_date"] = self._backup_start_timestamp.strftime("%Y-%m-%d")
        res["current_time"] = self._backup_start_timestamp.strftime("%H%M%S")

        res["lower"] = str.lower
        res["upper"] = str.upper

        return res

    def _get_additional_rendering_variables(self):
        targets_context = self.rendering_context.secrets_context["targets"][self.target_name]

        res = targets_context.get("additional_variables", {})
        self.logger.info("additional variables: {}".format(res))

        return res

    @property
    def config_context(self):
        return self.rendering_context.config_context

    @property
    def secrets_context(self):
        return self.rendering_context.secrets_context

    @property
    def flow_name(self):
        return self.rendering_context.flow_name

    @property
    def target_name(self):
        return self.rendering_context.target_name

    @property
    def backup_start_timestamp(self):
        return self._backup_start_timestamp



import os
import threading

from boto3.s3.transfer import MB, S3Transfer, TransferConfig
from botocore.exceptions import ClientError

from .base import WrongParameterTypeError


class ProgressPercentage(object):
    def __init__(self, logger, filename):
        self._filename = filename
        self._bytes_to_transmit = float(os.path.getsize(filename))
        self._transmitted_bytes_count = 0
        self._lock = threading.Lock()
        self._logger = logger

    def __call__(self, bytes_amount):
        with self._lock:
            self._transmitted_bytes_count += bytes_amount
            percentage = (self._transmitted_bytes_count / self._bytes_to_transmit) * 100
            percentage = round(percentage, 2)
            self._logger.info("transmitted: {}%".format(percentage))

    @staticmethod
    def _get_file_size(file_name):
        return os.path.getsize(file_name)


class S3BasicBotoClient(object):
    DEFAULT_TRANSMISSION_CHUNK_SIZE = 8 * MB
    DEFAULT_NOTIFICATION_THRESHHOLD = 1 * MB
    DEFAULT_TRANSMISSION_MAX_THREADS = 20
    DEFAULT_MAX_TRANSMISSION_ATTEMPTS = 5

    def __init__(self, logger, s3_client):
        self.logger = logger
        self._client = s3_client

    def create_bucket(self, bucket_name, region=None):
        try:
            if region is None:
                self._client.create_bucket(Bucket=bucket_name)
            else:
                location = {"LocationConstraint": region}
                self._client.create_bucket(
                    Bucket=bucket_name,
                    CreateBucketConfiguration=location
                )

        except ClientError as e:
            self.logger.error(e)
            return False

        return True

    def is_object_exists(self, bucket_name, object_name):
        try:
            self._client.head_object(Bucket=bucket_name, Key=object_name)
        except ClientError:
            return False

        return True

    def is_bucket_exists(self, bucket_name):
        try:
            _ = self._client.head_bucket(Bucket=bucket_name)  # noqa
        except ClientError as e:
            self.logger.debug(e)
            return False

        return True

    def upload_file(
        self,
        dest_bucket_name,
        dest_object_name,
        source_file_name,
        transfer_config=None
    ):
        if transfer_config is None:
            transfer_config = TransferConfig(
                multipart_threshold=S3BasicBotoClient.DEFAULT_NOTIFICATION_THRESHHOLD,
                max_concurrency=S3BasicBotoClient.DEFAULT_TRANSMISSION_MAX_THREADS,
                multipart_chunksize=S3BasicBotoClient.DEFAULT_TRANSMISSION_CHUNK_SIZE,
                num_download_attempts=S3BasicBotoClient.DEFAULT_MAX_TRANSMISSION_ATTEMPTS,
                use_threads=True
            )

        transfer = S3Transfer(self._client, config=transfer_config)

        transfer.upload_file(
            source_file_name,
            dest_bucket_name,
            dest_object_name,
            callback=ProgressPercentage(self.logger, source_file_name),
        )

        return True

    def copy_file_from_one_bucket_to_another(
        self,
        src_bucket_name,
        src_object_name,
        dest_bucket_name,
        dest_object_name,
    ):
        copy_source = {
            "Bucket": src_bucket_name,
            "Key": src_object_name
        }
        self._client.copy(copy_source, dest_bucket_name, dest_object_name)

    def put_object(self, dest_bucket_name, dest_object_name, src_data):
        """Add an object to an Amazon S3 bucket

        The src_data argument must be of type bytes or a string that references
        a file specification.

        :param dest_bucket_name: string
        :param dest_object_name: string
        :param src_data: bytes of data or string reference to file spec
        :return: True if src_data was added to dest_bucket/dest_object, otherwise
        False
        """

        object_data = None
        need_close = False
        try:
            if isinstance(src_data, bytes):
                object_data = src_data
            elif isinstance(src_data, str):
                need_close = True
                object_data = open(src_data, "rb")
            else:
                msg = "Type of {} for the argument 'src_data' is not supported.".format(str(type(src_data)))
                self.logger.error(msg)
                raise WrongParameterTypeError(msg)

            self._put_object(dest_bucket_name, dest_object_name, object_data)

        finally:
            if need_close:
                object_data.close()

    def list_files_in_folder(self, bucket_name, folder=""):
        response = self._client.list_objects(Bucket=bucket_name, Prefix=folder)
        return [content.get("Key") for content in response.get("Contents", [])]

    def delete_object(self, bucket_name, key):
        self._client.delete_object(Bucket=bucket_name, Key=key)

    def get_object_tags(self, bucket_name, key):
        ret = {}

        resp = self._client.get_object_tagging(Bucket=bucket_name, Key=key)
        if "TagSet" not in resp:
            return ret

        tags_set = resp["TagSet"]
        for tags_set_item in tags_set:
            ret[tags_set_item["Key"]] = tags_set_item["Value"]

        return ret

    def set_object_tags(self, bucket_name, key, tags):
        tags_list = [{"Key": str(key), "Value": str(value)} for key, value in tags.items()]
        self._client.put_object_tagging(Bucket=bucket_name, Key=key, Tagging={"TagSet": tags_list})

    def delete_object_tags(self, bucket_name, key):
        self._client.get_object_tagging(Bucket=bucket_name, Key=key)

    def _put_object(self, dest_bucket_name, dest_object_name, object_data):
        # Put the object
        try:
            self._client.put_object(Bucket=dest_bucket_name, Key=dest_object_name, Body=object_data)
        except Exception as e:
            # AllAccessDisabled error == bucket not found
            # NoSuchKey or InvalidRequest error == (dest bucket/obj == src bucket/obj)
            self.logger.error(e)
            raise



from .step_calculate_file_hash_and_save_to_file import StepCalculateFileHashAndSaveToFile
from .step_compress_file_with_7z import StepCompressFileWith7Z
from .step_make_directory_for_backup import StepMakeDirectoryForBackup
from .step_make_firebird_database_backup import StepMakeFirebirdDatabaseBackup
from .step_make_pg_win_database_backup import StepMakePgDatabaseWinBackup
from .step_s3_multipart_upload_with_rotation import StepS3MultipartUploadWithRotation
from .step_s3_strict_uploader import StepS3StrictUploader
from .step_validate_7z_archive import StepValidate7ZArchive


class StepsFactory(object):
    def __init__(self):
        self._known_steps = dict()
        pass

    def register_class(self, cls):
        self._known_steps[cls.step_name()] = cls

    def create_object(self, step_name, **kwargs):
        return self._known_steps[step_name](**kwargs)

    def is_step_known(self, step_name):
        return step_name in self._known_steps


def create_steps_factory():
    factory = StepsFactory()

    factory.register_class(StepMakeDirectoryForBackup)
    factory.register_class(StepMakeFirebirdDatabaseBackup)
    factory.register_class(StepCalculateFileHashAndSaveToFile)
    factory.register_class(StepCompressFileWith7Z)
    factory.register_class(StepValidate7ZArchive)
    factory.register_class(StepS3MultipartUploadWithRotation)
    factory.register_class(StepS3StrictUploader)
    factory.register_class(StepMakePgDatabaseWinBackup)

    return factory



from setuptools import find_packages, setup  # noqa
from yabtool import __version__  # noqa

try:  # for pip >= 10
    from pip._internal.req import parse_requirements  # noqa
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements  # noqa


PACKAGE_NAME = "yabtool"

README_FILE = "README.md"
install_reqs = parse_requirements("requirements.txt", session=False)

requirements = [str(ir.requirement) for ir in install_reqs
                if ir and ir.requirement and str(ir.requirement).strip()]
hyphen_package_name = PACKAGE_NAME.replace("_", "-")
LONG_DESCTIPTION = "Please read https://github.com/JFF-Bohdan/yabtool/blob/master/README.md for information"


def read_file_content(file_name):
    with open(file_name) as f:
        return f.read()


if __name__ == "__main__":
    packages_to_remove = ["tests"]
    packages = find_packages()

    for item in packages_to_remove:
        if item in packages:
            packages.remove(item)

    setup(
        name=PACKAGE_NAME,
        packages=packages,
        package_dir={
            PACKAGE_NAME: PACKAGE_NAME
        },
        package_data={
            PACKAGE_NAME: ["./config/*.yaml"]
        },
        version=__version__,
        description="Module to support VRC-T70 hardware",
        long_description=LONG_DESCTIPTION,
        long_description_content_type="text/x-rst",
        author="Bohdan Danishevsky",
        author_email="dbn@aminis.com.ua",
        url="https://github.com/JFF-Bohdan/yabtool",
        keywords=["flexible backup tool", "configurable multipurpose backup tool", "backup tool"],
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        install_requires=requirements,
        classifiers=[],
        license="MIT",
        zip_safe=False
    )



import argparse
import os
import shutil
import sys

import loguru
from loguru._defaults import LOGURU_FORMAT
from yabtool.supported_notifications.email_notifications import EmailRenderer, EmailSender
from yabtool.version import __version__


from .yabtool_flow_orchestrator import YabtoolFlowOrchestrator


def get_cli_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version="%(prog)s {}".format(__version__)
    )

    parser.add_argument(
        "--log-level",
        "-l",
        action="store",
        default="INFO",
        help="Specify log level (DEBUG, INFO, ...)"
    )

    parser.add_argument(
        "--disable-voting",
        "-k",
        action="store_true",
        help="Allow voting for flow execution skipping"
    )

    parser.add_argument(
        "--secrets",
        "-s",
        required=True,
        action="store",
        help="Path to file with secrets"
    )

    parser.add_argument(
        "--config",
        "-c",
        action="store",
        help="Path to main configuration file"
    )

    parser.add_argument(
        "--dry-run",
        "-y",
        action="store_true",
        default=None,
        help="Perform dry run only"
    )

    parser.add_argument(
        "--target",
        "-d",
        action="store",
        help="Specify target in secret file (overrides default in secrets file if specified)"
    )

    parser.add_argument(
        "--temporary-folder",
        "-t",
        action="store",
        help="Path to temporary folder that will be used by yabtool"
    )

    parser.add_argument(
        "--flow",
        "-f",
        action="store",
        help="Required flow name"
    )

    parser.add_argument(
        "--add-main-log",
        "-m",
        action="store_true",
        default=False,
        help="Add main log to logs folder"
    )

    parser.add_argument(
        "--add-session-log",
        "-p",
        action="store_true",
        default=False,
        help="Add session log file"
    )

    return parser.parse_known_args()


class YabtoolApplication(object):
    def __init__(self):
        self.logger = None
        self.rendering_context = None
        self._session_log_path = None

    def run(self):
        args, unknown_args = get_cli_args()
        self._initialize_logger(args)
        self.logger.debug(f"Unknown command line arguments: {unknown_args}")

        flow_orchestrator = YabtoolFlowOrchestrator(self.logger)

        if args.disable_voting:
            flow_orchestrator.skip_voting_enabled = False

        only_dry_run = None
        folder_name = None
        try:
            flow_orchestrator.initialize(args, unknown_args)

            folder_name = flow_orchestrator.rendering_context.temporary_folder
            root_temporary_folder = flow_orchestrator.rendering_context.root_temporary_folder
            flow_name = flow_orchestrator.flow_name

            if args.add_main_log:
                main_logs_folder = os.path.join(root_temporary_folder, "logs", "main")
                self._add_main_log(main_logs_folder, args)

            if args.add_session_log:
                session_logs_folder = os.path.join(root_temporary_folder, "logs", "session")
                self._add_session_log(session_logs_folder, flow_orchestrator.backup_start_timestamp, args)

            flow_orchestrator.dry_run()
            self.logger.info("dry run for flow '{}' performed".format(flow_name))

            only_dry_run = True
            if not args.dry_run:
                self.logger.info("flow '{}' started".format(flow_name))
                flow_orchestrator.run()
                only_dry_run = False

            flow_orchestrator.print_stat()
            self._send_notifications(flow_orchestrator, only_dry_run=only_dry_run)

        except BaseException as e:
            self.logger.exception("Error performing flow. Exception: {}".format(e))
            self._send_notifications(flow_orchestrator, succeeded=False, exception=e, only_dry_run=only_dry_run)

        finally:
            if flow_orchestrator.rendering_context.remove_temporary_folder and folder_name:
                if folder_name and os.path.exists(folder_name) and os.path.isdir(folder_name):
                    self.logger.info("going to remove temporary folder: {}".format(folder_name))
                    self._remove_temporary_folder(folder_name)

            else:
                self.logger.info("output folder removal disabled. folder name: '{}'".format(folder_name))

        return True

    def _send_notifications(self, flow_orchestrator, succeeded=True, exception=None, only_dry_run=False):
        enabled_notifications = self._get_enabled_notifications(flow_orchestrator)
        if not enabled_notifications:
            self.logger.info("no enabled notifications")
            return

        self.logger.debug("sending notifications for rules: {}".format(enabled_notifications))
        for notification_type, notification_data in enabled_notifications:
            if notification_type == "email":
                email_notifier = EmailRenderer(self.logger)

                email_notifier.flow_orchestrator = flow_orchestrator
                email_notifier.notification_data = notification_data
                email_notifier.succeeded = succeeded
                email_notifier.exception = exception
                email_notifier.only_dry_run = only_dry_run

                rendered_data = email_notifier.render()
                if self._session_log_path:
                    rendered_data.attachments.append(self._session_log_path)

                sender = EmailSender(self.logger, notification_data)
                sender.send(rendered_data)

            else:
                self.logger.error("unsupported notification type: '{}'".format(notification_type))

    def _get_enabled_notifications(self, flow_orchestrator):
        res = []

        target_data = flow_orchestrator.secrets_context["targets"][flow_orchestrator.rendering_context.target_name]
        notifications = target_data.get("notifications")
        if not notifications:
            return res

        for notification_type, notification in notifications.items():
            self.logger.debug("validating notification type for '{}'".format(notification_type))
            if not self._is_known_notification_type(notification_type):
                self.logger.warning("unknown notification type for '{}'".format(notification_type))
                continue

            if notification.get("enabled", False):
                notification_item = (notification_type, notification)
                res.append(notification_item)

        return res

    def _is_known_notification_type(self, notification_type):
        return notification_type in ["email"]

    def _initialize_logger(self, args):
        self.logger = loguru.logger
        self.logger.remove()
        self.logger.add(sys.stdout, format=LOGURU_FORMAT, level=args.log_level)

    def _add_main_log(self, logs_folder, args):
        if not os.path.exists(logs_folder):
            os.makedirs(logs_folder)

        path = os.path.join(logs_folder, "main_log.log")
        self.logger.add(path, rotation="10 Mb", retention="60 days", compression="zip", level=args.log_level)

    def _add_session_log(self, logs_folder, timestamp_begin, args):
        if not os.path.exists(logs_folder):
            os.makedirs(logs_folder)

        session_log_suffix = timestamp_begin.strftime("%Y-%m-%dT%H%M%S")
        path = os.path.join(logs_folder, "session_{}.log".format(session_log_suffix))
        self._session_log_path = path
        self.logger.add(path, level=args.log_level)

    def _remove_temporary_folder(self, folder_name):
        try:
            shutil.rmtree(folder_name)
        except Exception as e:
            self.logger.error("error removing folder '{}': {}".format(folder_name, e))
